@@ -55,6 +55,7 @@
 #define KEY_GNOME_ANIMATIONS "enable-animations"
 #define KEY_GNOME_CURSOR_THEME "cursor-theme"
 #define KEY_GNOME_CURSOR_SIZE "cursor-size"
+#define KEY_XKB_OPTIONS "xkb-options"
 
 #define KEY_OVERLAY_KEY "overlay-key"
 #define KEY_WORKSPACES_ONLY_ON_PRIMARY "workspaces-only-on-primary"
@@ -65,6 +66,7 @@
 #define SCHEMA_GENERAL         "org.gnome.desktop.wm.preferences"
 #define SCHEMA_MUTTER          "org.gnome.mutter"
 #define SCHEMA_INTERFACE       "org.gnome.desktop.interface"
+#define SCHEMA_INPUT_SOURCES   "org.gnome.desktop.input-sources"
 
 #define SETTINGS(s) g_hash_table_lookup (settings_schemas, (s))
 
@@ -115,6 +117,7 @@ static gboolean workspaces_only_on_primary = FALSE;
 
 static gboolean no_tab_popup = FALSE;
 
+static char *iso_next_group_option = NULL;
 
 static void handle_preference_update_enum (GSettings *settings,
                                            gchar     *key);
@@ -122,7 +125,6 @@ static gboolean update_binding         (MetaKeyPref *binding,
                                         gchar      **strokes);
 static gboolean update_key_binding     (const char  *key,
                                         gchar      **strokes);
-static gboolean update_workspace_names (void);
 
 static void settings_changed (GSettings      *settings,
                               gchar          *key,
@@ -140,11 +142,11 @@ static gboolean theme_name_handler (GVariant*, gpointer*, gpointer);
 static gboolean mouse_button_mods_handler (GVariant*, gpointer*, gpointer);
 static gboolean button_layout_handler (GVariant*, gpointer*, gpointer);
 static gboolean overlay_key_handler (GVariant*, gpointer*, gpointer);
+static gboolean iso_next_group_handler (GVariant*, gpointer*, gpointer);
 
 static void     do_override               (char *key, char *schema);
 
 static void     init_bindings             (void);
-static void     init_workspace_names      (void);
 
 
 typedef struct
@@ -198,6 +200,13 @@ typedef struct
   gchar **target;
 } MetaStringPreference;
 
+typedef struct
+{
+  MetaBasePreference base;
+  GSettingsGetMapping handler;
+  gchar ***target;
+} MetaStringArrayPreference;
+
 typedef struct
 {
   MetaBasePreference base;
@@ -436,6 +445,27 @@ static MetaStringPreference preferences_string[] =
     { { NULL, 0, 0 }, NULL },
   };
 
+static MetaStringArrayPreference preferences_string_array[] =
+  {
+    {
+      { KEY_WORKSPACE_NAMES,
+        SCHEMA_GENERAL,
+        META_PREF_KEYBINDINGS,
+      },
+      NULL,
+      &workspace_names,
+    },
+    {
+      { KEY_XKB_OPTIONS,
+        SCHEMA_INPUT_SOURCES,
+        META_PREF_KEYBINDINGS,
+      },
+      iso_next_group_handler,
+      NULL,
+    },
+    { { NULL, 0, 0 }, NULL },
+  };
+
 static MetaIntPreference preferences_int[] =
   {
     {
@@ -555,6 +585,42 @@ handle_preference_init_string (void)
     }
 }
 
+static void
+handle_preference_init_string_array (void)
+{
+  MetaStringArrayPreference *cursor = preferences_string_array;
+
+  while (cursor->base.key != NULL)
+    {
+      char **value;
+
+      /* Complex keys have a mapping function to check validity */
+      if (cursor->handler)
+        {
+          if (cursor->target)
+            meta_bug ("%s has both a target and a handler\n", cursor->base.key);
+
+          g_settings_get_mapped (SETTINGS (cursor->base.schema),
+                                 cursor->base.key, cursor->handler, NULL);
+        }
+      else
+        {
+          if (!cursor->target)
+            meta_bug ("%s must have handler or target\n", cursor->base.key);
+
+          if (*(cursor->target))
+            g_strfreev (*(cursor->target));
+
+          value = g_settings_get_strv (SETTINGS (cursor->base.schema),
+                                       cursor->base.key);
+
+          *(cursor->target) = value;
+        }
+
+      ++cursor;
+    }
+}
+
 static void
 handle_preference_init_int (void)
 {
@@ -673,6 +739,56 @@ handle_preference_update_string (GSettings *settings,
     queue_changed (cursor->base.pref);
 }
 
+static void
+handle_preference_update_string_array (GSettings *settings,
+                                       gchar *key)
+{
+  MetaStringArrayPreference *cursor = preferences_string_array;
+  gboolean inform_listeners = FALSE;
+
+  while (cursor->base.key != NULL && strcmp (key, cursor->base.key) != 0)
+    ++cursor;
+
+  if (cursor->base.key==NULL)
+    /* Didn't recognise that key. */
+    return;
+
+  /* Complex keys have a mapping function to check validity */
+  if (cursor->handler)
+    {
+      if (cursor->target)
+        meta_bug ("%s has both a target and a handler\n", cursor->base.key);
+
+      g_settings_get_mapped (SETTINGS (cursor->base.schema),
+                             cursor->base.key, cursor->handler, NULL);
+    }
+  else
+    {
+      char **values, **previous;
+      int n_values, n_previous, i;
+
+      if (!cursor->target)
+        meta_bug ("%s must have handler or target\n", cursor->base.key);
+
+      values = g_settings_get_strv (SETTINGS (cursor->base.schema),
+                                    cursor->base.key);
+      n_values = g_strv_length (values);
+      previous = *(cursor->target);
+      n_previous = previous ? g_strv_length (previous) : 0;
+
+      inform_listeners = n_previous != n_values;
+      for (i = 0; i < n_values && !inform_listeners; i++)
+        inform_listeners = g_strcmp0 (values[i], previous[i]) != 0;
+
+      if (*(cursor->target))
+        g_strfreev (*(cursor->target));
+      *(cursor->target) = values;
+    }
+
+  if (inform_listeners)
+    queue_changed (cursor->base.pref);
+}
+
 static void
 handle_preference_update_int (GSettings *settings,
                               gchar *key)
@@ -857,6 +973,11 @@ meta_prefs_init (void)
                     G_CALLBACK (settings_changed), NULL);
   g_hash_table_insert (settings_schemas, g_strdup (SCHEMA_INTERFACE), settings);
 
+  settings = g_settings_new (SCHEMA_INPUT_SOURCES);
+  g_signal_connect (settings, "changed::" KEY_XKB_OPTIONS,
+                    G_CALLBACK (settings_changed), NULL);
+  g_hash_table_insert (settings_schemas, g_strdup (SCHEMA_INPUT_SOURCES), settings);
+
 
   for (tmp = overridden_keys; tmp; tmp = tmp->next)
     {
@@ -869,10 +990,10 @@ meta_prefs_init (void)
   handle_preference_init_enum ();
   handle_preference_init_bool ();
   handle_preference_init_string ();
+  handle_preference_init_string_array ();
   handle_preference_init_int ();
 
   init_bindings ();
-  init_workspace_names ();
 }
 
 static gboolean
@@ -1020,14 +1141,6 @@ settings_changed (GSettings *settings,
   MetaEnumPreference *cursor;
   gboolean found_enum;
 
-  /* String array, handled separately */
-  if (strcmp (key, KEY_WORKSPACE_NAMES) == 0)
-    {
-      if (update_workspace_names ())
-        queue_changed (META_PREF_WORKSPACE_NAMES);
-      return;
-    }
-
   value = g_settings_get_value (settings, key);
   type = g_variant_get_type (value);
 
@@ -1035,6 +1148,8 @@ settings_changed (GSettings *settings,
     handle_preference_update_bool (settings, key);
   else if (g_variant_type_equal (type, G_VARIANT_TYPE_INT32))
     handle_preference_update_int (settings, key);
+  else if (g_variant_type_equal (type, G_VARIANT_TYPE_STRING_ARRAY))
+    handle_preference_update_string_array (settings, key);
   else if (g_variant_type_equal (type, G_VARIANT_TYPE_STRING))
     {
       cursor = preferences_enum;
@@ -1554,6 +1669,39 @@ overlay_key_handler (GVariant *value,
   return TRUE;
 }
 
+static gboolean
+iso_next_group_handler (GVariant *value,
+                        gpointer *result,
+                        gpointer  data)
+{
+  const char **xkb_options, **p;
+  const char *option = NULL;
+  gboolean changed = FALSE;
+
+  *result = NULL; /* ignored */
+  xkb_options = g_variant_get_strv (value, NULL);
+
+  for (p = xkb_options; p && *p; ++p)
+    if (g_str_has_prefix (*p, "grp:"))
+      {
+        option = (*p + 4);
+        break;
+      }
+
+  changed = (g_strcmp0 (option, iso_next_group_option) != 0);
+
+  if (changed)
+    {
+      g_free (iso_next_group_option);
+      iso_next_group_option = g_strdup (option);
+      queue_changed (META_PREF_KEYBINDINGS);
+    }
+
+  g_free (xkb_options);
+
+  return TRUE;
+}
+
 const PangoFontDescription*
 meta_prefs_get_titlebar_font (void)
 {
@@ -1748,20 +1896,15 @@ init_bindings (void)
   g_hash_table_insert (key_bindings, g_strdup ("overlay-key"), pref);
 }
 
-static void
-init_workspace_names (void)
-{
-  update_workspace_names ();
-}
-
 static gboolean
 update_binding (MetaKeyPref *binding,
                 gchar      **strokes)
 {
+  GSList *old_bindings, *a, *b;
+  gboolean changed;
   unsigned int keysym;
   unsigned int keycode;
   MetaVirtualModifier mods;
-  gboolean changed = FALSE;
   MetaKeyCombo *combo;
   int i;
 
@@ -1769,13 +1912,9 @@ update_binding (MetaKeyPref *binding,
               "Binding \"%s\" has new GSettings value\n",
               binding->name);
 
-  /* Okay, so, we're about to provide a new list of key combos for this
-   * action. Delete any pre-existing list.
-   */
-  g_slist_foreach (binding->bindings, (GFunc) g_free, NULL);
-  g_slist_free (binding->bindings);
+  old_bindings = binding->bindings;
   binding->bindings = NULL;
-  
+
   for (i = 0; strokes && strokes[i]; i++)
     {
       keysym = 0;
@@ -1810,8 +1949,6 @@ update_binding (MetaKeyPref *binding,
            * Changing the key in response to a modification could lead to cyclic calls. */
           continue;
         }
-  
-      changed = TRUE;
 
       combo = g_malloc0 (sizeof (MetaKeyCombo));
       combo->keysym = keysym;
@@ -1826,6 +1963,34 @@ update_binding (MetaKeyPref *binding,
 
   binding->bindings = g_slist_reverse (binding->bindings);
 
+  a = old_bindings;
+  b = binding->bindings;
+  while (TRUE)
+    {
+      if ((!a && b) || (a && !b))
+        {
+          changed = TRUE;
+          break;
+        }
+      else if (!a && !b)
+        {
+          changed = FALSE;
+          break;
+        }
+      else if (memcmp (a->data, b->data, sizeof (MetaKeyCombo)) != 0)
+        {
+          changed = TRUE;
+          break;
+        }
+      else
+        {
+          a = a->next;
+          b = b->next;
+        }
+    }
+
+  g_slist_free_full (old_bindings, g_free);
+
   return changed;
 }
 
@@ -1841,41 +2006,6 @@ update_key_binding (const char *key,
     return FALSE;
 }
 
-static gboolean
-update_workspace_names (void)
-{
-  int i;
-  char **names;
-  int n_workspace_names, n_names;
-  gboolean changed = FALSE;
-
-  names = g_settings_get_strv (SETTINGS (SCHEMA_GENERAL), KEY_WORKSPACE_NAMES);
-  n_names = g_strv_length (names);
-  n_workspace_names = workspace_names ? g_strv_length (workspace_names) : 0;
-
-  for (i = 0; i < n_names; i++)
-    if (n_workspace_names < i + 1 || !workspace_names[i] ||
-        g_strcmp0 (names[i], workspace_names[i]) != 0)
-      {
-        changed = TRUE;
-        break;
-      }
-
-  if (n_workspace_names != n_names)
-    changed = TRUE;
-
-  if (changed)
-    {
-      if (workspace_names)
-        g_strfreev (workspace_names);
-      workspace_names = names;
-    }
-  else
-    g_strfreev (names);
-
-  return changed;
-}
-
 const char*
 meta_prefs_get_workspace_name (int i)
 {
@@ -2075,6 +2205,12 @@ meta_prefs_get_overlay_binding (MetaKeyCombo *combo)
   *combo = overlay_key_combo;
 }
 
+const char *
+meta_prefs_get_iso_next_group_option (void)
+{
+  return iso_next_group_option;
+}
+
 GDesktopTitlebarAction
 meta_prefs_get_action_double_click_titlebar (void)
 {
